# -*- coding: utf-8 -*-
from datetime import datetime
from calendar import timegm
from rest_framework_jwt.settings import api_settings as jwt_api_settings

from .settings import api_settings


def jwt_get_username_from_payload_handler(payload):
    return payload['uid']


def jwt_payload_handler(user):
    # store only user_id in the payload
    payload = {
        'uid': user.id,
        'exp': datetime.utcnow() + jwt_api_settings.JWT_EXPIRATION_DELTA
    }
    if jwt_api_settings.JWT_ALLOW_REFRESH:
        payload['orig_iat'] = timegm(datetime.utcnow().utctimetuple())
    if jwt_api_settings.JWT_AUDIENCE is not None:
        payload['aud'] = jwt_api_settings.JWT_AUDIENCE

    if jwt_api_settings.JWT_ISSUER is not None:
        payload['iss'] = jwt_api_settings.JWT_ISSUER
    return payload


def jwt_response_payload_handler(token, user=None, request=None):
    UserSerializer = api_settings.USER_SERIALIZER
    return {
        'token': token,
        'user': UserSerializer(user, context={'request': request}).data
    }



# -*- coding: utf-8 -*-
import datetime
from django.contrib.auth import get_user_model, update_session_auth_hash
from django.core.signing import BadSignature, SignatureExpired
from django.db import transaction
from django.utils.translation import ugettext as _
from rest_framework import status
from rest_framework.exceptions import NotFound, APIException
from rest_framework.generics import GenericAPIView
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework_jwt.settings import api_settings as jwt_api_settings
from rest_framework_jwt.views import RefreshJSONWebToken, ObtainJSONWebToken

from .email import send_confirmation_email, resend_email, send_passwordreset_email
from .serializers import RegisterSerializer, PasswordSerializer, EmailSerializer
from .serializers import RefreshJSONWebTokenSerializer, ObtainJSONWebTokenSerializer
from .settings import api_settings
from .signing import (
    verify_email_confirm_token, create_email_confirm_token, create_password_reset_token,
    verify_activation_token, verify_password_reset_token, verify_email_change_token)


jwt_payload_handler = jwt_api_settings.JWT_PAYLOAD_HANDLER
jwt_encode_handler = jwt_api_settings.JWT_ENCODE_HANDLER
jwt_response_payload_handler = jwt_api_settings.JWT_RESPONSE_PAYLOAD_HANDLER
refresh_jwt_token = RefreshJSONWebToken.as_view(serializer_class=RefreshJSONWebTokenSerializer)
obtain_jwt_token = ObtainJSONWebToken.as_view(serializer_class=ObtainJSONWebTokenSerializer)


def create_jwt_response(request, user, statuscode=status.HTTP_201_CREATED):
    token = jwt_encode_handler(jwt_payload_handler(user))
    res = jwt_response_payload_handler(token, user, request)
    return Response(res, status=statuscode)


class TokenExpired(APIException):
    status_code = status.HTTP_410_GONE
    default_detail = _("Token has expired.")


class UserAlreadyConfirmed(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _("User has already been confirmed.")


class UserAlreadyActivated(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _("User has already been activated.")


class UserNotMached(APIException):
    status_code = status.HTTP_409_CONFLICT
    default_detail = _("User did not match.")


class RegisterView(GenericAPIView):
    """
    emailとpasswordを指定してユーザ登録
    """
    serializer_class = RegisterSerializer
    permission_classes = (AllowAny,)

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        s = self.get_serializer(data=request.data)
        s.is_valid(raise_exception=True)
        # email confirmationが必要ならis_activeをfalseで作成する
        user = s.save(is_active=not(api_settings.REQUIRE_EMAIL_CONFIRMATION))
        if api_settings.REQUIRE_EMAIL_CONFIRMATION:
            token = create_email_confirm_token(user)
            send_confirmation_email(user, token)
            return Response()
        response = create_jwt_response(request, user)
        if jwt_api_settings.JWT_AUTH_COOKIE:
            expiration = (datetime.utcnow() + jwt_api_settings.JWT_EXPIRATION_DELTA)
            response.set_cookie(jwt_api_settings.JWT_AUTH_COOKIE,
                                token,
                                expires=expiration,
                                httponly=True)
        return response


class EmailConfirmView(APIView):
    """
    emailの存在チェックをするview
    これを通らないとuserがactiveにならない->作成時にis_activeをfalseにするとかする
    """
    permission_classes = (AllowAny,)

    def verify_token(self, token):
        try:
            user = verify_email_confirm_token(token, datetime.timedelta(hours=24))
            if user is None:
                raise NotFound()
            if user.is_active:
                raise UserAlreadyConfirmed()
            return user
        except SignatureExpired:
            raise TokenExpired()
        except BadSignature:
            raise NotFound()

    def get(self, request, token):
        self.verify_token(token)
        return Response()  # emailも返す?

    def post(self, request, token):
        user = self.verify_token(token)
        user.is_active = True
        user.save()
        return create_jwt_response(request, user, statuscode=status.HTTP_200_OK)


class EmailChangeConfirmView(APIView):
    """
    新しいemailの存在チェックをするview
    これがアクセスされないとユーザのemailを変更しない
    ToDo: 保存時に既にemailが衝突している場合は対応しない
    """

    def verify_email_token(self, token, request):
        try:
            user, email = verify_email_change_token(token, datetime.timedelta(hours=24))
            if user is None:
                raise NotFound()
            if user.id != request.user.id:
                raise UserNotMached()
            return user, email
        except SignatureExpired:
            raise TokenExpired()
        except BadSignature:
            raise NotFound()

    def get(self, request, token):
        _, email = self.verify_email_token(token, request)
        return Response({"email": email})  # emailも返す?

    def post(self, request, token):
        user, email = self.verify_email_token(token, request)
        user.email = email
        user.save()
        return create_jwt_response(request, user, statuscode=status.HTTP_200_OK)


class ActivationView(GenericAPIView):
    """
    userが他社によって登録されたので、パスワードを設定してactivationする
    """
    serializer_class = api_settings.ACTIVATION_SERIALIZER
    permission_classes = (AllowAny,)
    queryset = get_user_model().objects.none()

    def verify_token(self, token):
        try:
            user = verify_activation_token(token, datetime.timedelta(hours=24))
            if user is None:
                raise NotFound()
            if user.is_active:
                raise UserAlreadyActivated()
            return user
        except SignatureExpired:
            raise TokenExpired()
        except BadSignature:
            raise NotFound()

    def get(self, request, token):
        u = self.verify_token(token)
        return Response({"email": u.email})

    def post(self, request, token):
        user = self.verify_token(token)
        s = self.get_serializer(data=request.data)
        s.is_valid(raise_exception=True)
        user.is_active = True
        user.set_password(s.password)
        user.save()
        return create_jwt_response(request, user, statuscode=status.HTTP_200_OK)


class PasswordResetRequestView(GenericAPIView):
    permission_classes = (AllowAny,)
    serializer_class = EmailSerializer
    queryset = get_user_model().objects.none()

    def post(self, request):
        s = self.get_serializer(data=request.data)
        s.is_valid(raise_exception=True)
        users = get_user_model().objects.filter(email=s.validated_data['email'])
        for user in users:
            token = create_password_reset_token(user)
            send_passwordreset_email(user, token)
        return Response()


class PasswordResetView(GenericAPIView):
    serializer_class = PasswordSerializer
    permission_classes = (AllowAny,)

    def verify_token(self, token):
        try:
            user = verify_password_reset_token(token, datetime.timedelta(hours=24))
            if user is None:
                raise NotFound()
            return user
        except SignatureExpired:
            raise TokenExpired()
        except BadSignature:
            raise NotFound()

    def get(self, request, token):
        u = self.verify_token(token)
        return Response({"email": u.email})

    def post(self, request, token):
        user = self.verify_token(token)
        s = self.get_serializer(data=request.data)
        s.is_valid(raise_exception=True)
        user.is_active = True
        user.set_password(s.password)
        user.save()
        return Response()  # resetの場合は再度loginさせる方が良い気はする


class ResendTokenView(APIView):
    permission_classes = (AllowAny,)

    def post(self, request, token):
        try:
            resend_email(token)
        except BadSignature:
            raise NotFound()
        return Response()  # emailを遅れたら空の200を返す


class PasswordChangeView(GenericAPIView):
    serializer_class = PasswordSerializer

    def post(self, request):
        s = self.get_serializer(data=request.data)
        s.is_valid(raise_exception=True)
        request.user.set_password(s.password)
        request.user.save()
        # 本番では消したほうが良いかもしれない
        update_session_auth_hash(request, request.user)
        return Response()  # 正しくパスワード変更できたら空の200を返す



from django.core.mail import EmailMessage, EmailMultiAlternatives
from django.template import TemplateDoesNotExist
from django.template.loader import render_to_string

from .settings import api_settings
from .signing import create_email_confirmation_url, create_activation_url, create_password_reset_url
from .signing import get_user_and_tokentype, get_email_from_token
from .signing import TOKEN_EMAIL_CONFIRM, TOKEN_ACTIVATION, TOKEN_PASSWORD_RESET, TOKEN_EMAIL_CHANGE
from. signing import create_email_confirm_token, create_activation_token, create_password_reset_token
from. signing import create_email_change_url, create_email_change_token


def resend_email(old_token):
    user, token_type = get_user_and_tokentype(old_token)
    if token_type == TOKEN_EMAIL_CONFIRM:
        send_confirmation_email(user, create_email_confirm_token(user))
    elif token_type == TOKEN_ACTIVATION:
        send_activation_email(user, create_activation_token(user))
    elif token_type == TOKEN_PASSWORD_RESET:
        send_passwordreset_email(user, create_password_reset_token(user))
    elif token_type == TOKEN_EMAIL_CHANGE:
        email = get_email_from_token(old_token)
        send_email_change_email(email, create_email_change_token(user, email))


def send_confirmation_email(user, token):
    context = {"token_url": create_email_confirmation_url(token)}
    template_prefix = "drf_jwt_util/email/confirmation"
    send_email(user.email, template_prefix, context)


def send_email_change_email(email, token):
    context = {"token_url": create_email_change_url(token)}
    template_prefix = "drf_jwt_util/email/emailchange"
    send_email(email, template_prefix, context)


def send_activation_email(user, token):
    context = {"token_url": create_activation_url(token)}
    template_prefix = "drf_jwt_util/email/activation"
    send_email(user.email, template_prefix, context)


def send_passwordreset_email(user, token):
    context = {"token_url": create_password_reset_url(token)}
    template_prefix = "drf_jwt_util/email/passwordreset"
    send_email(user.email, template_prefix, context)


def send_email(to_email, template_prefix, context):
    def render_email(template, context):
        try:
            return render_to_string(template, context).strip()
        except TemplateDoesNotExist:
            return None
    from_email = api_settings.FROM_EMAIL
    subject = render_to_string('{}_subject.txt'.format(template_prefix), context)
    subject = " ".join(subject.splitlines()).strip()

    html_rendered = render_email('{}_message.html'.format(template_prefix))
    text_rendered = render_email('{}_message.txt'.format(template_prefix))
    if not html_rendered and not text_rendered:
        raise TemplateDoesNotExist()
    if text_rendered:
        msg = EmailMultiAlternatives(subject, text_rendered, from_email, [to_email])
        if html_rendered:
            msg.attach_alternative(html_rendered, 'text/html')
    else:
        msg = EmailMessage(subject, html_rendered, from_email, [to_email])
        msg.content_subtype = 'html'
    return msg.send()



from rest_framework import serializers


class SocialSerializer(serializers.Serializer):
    provider = serializers.CharField(max_length=255, required=True)
    access_token = serializers.CharField(max_length=4096, required=True, trim_whitespace=True)



from django.conf import settings
from rest_framework.settings import APISettings


USER_SETTINGS = getattr(settings, 'DRF_JWT_UTIL', None)

DEFAULTS = {
    'REQUIRE_EMAIL_CONFIRMATION': False,
    "FRONTEND_DOMAIN_WITH_SCHEME": "http://127.0.0.1:3000",
    "PASSWORD_RESET_PATH_TEMPLATE": "/reset/{token}",
    "ACTIVATION_PATH_TEMPLATE": "/activate/{token}",
    "EMAIL_CONFIRMATION_PATH_TEMPLATE": "/confirm/{token}",
    "EMAIL_CHANGE_PATH_TEMPLATE": "/emailchange/{token}",
    "USER_SERIALIZER": "drf_jwt_util.serializers.DefaultUserSerializer",
    "ACTIVATION_SERIALIZER": "drf_jwt_util.serializers.PasswordSerializer",
    "FROM_EMAIL": settings.DEFAULT_FROM_EMAIL,
}


IMPORT_STRINGS = [
    "USER_SERIALIZER",
    "ACTIVATION_SERIALIZER",
]

api_settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)



from rest_framework import exceptions
from rest_framework_jwt.authentication import JSONWebTokenAuthentication
from django.contrib.auth import get_user_model
from django.utils.translation import ugettext as _


class JWTAuthentication(JSONWebTokenAuthentication):
    def authenticate_credentials(self, payload):
        user = self.get_user_from_payload(payload)
        return user

    def get_user_from_payload(self, payload):
        User = get_user_model()
        try:
            user = User.objects.get(pk=payload.get('uid'))
        except User.DoesNotExist:
            msg = _('Invalid signature.')
            raise exceptions.AuthenticationFailed(msg)
        if not user.is_active:
            msg = _('User account is disabled.')
            raise exceptions.AuthenticationFailed(msg)
        return user



# -*- coding: utf-8 -*-
from django.contrib.auth import get_user_model
from django.core import signing
from django.core.signing import BadSignature, SignatureExpired

from .settings import api_settings


TOKEN_EMAIL_CONFIRM = "confirm"
TOKEN_EMAIL_CHANGE = "change"
TOKEN_ACTIVATION = "activation"
TOKEN_PASSWORD_RESET = "reset"


def create_token(user, type, compress=True):
    return signing.dumps({"u": user.id, "type": type}, compress=compress)


def create_email_token(user, email, type, compress=True):
    return signing.dumps({"u": user.id, "email": email, "type": type}, compress=compress)


def check_type(obj, type):
    if "type" in obj and obj["type"] == type:
        return True
    else:
        return False


def verify_token(token, type, max_age):
    # BadSignatureかSinatureExpiredが例外として返される
    # それ以外は、userの情報が無い、userが存在しない場合はNoneを返し、いる場合はuserを返す
    try:
        obj = signing.loads(token, max_age=max_age)
        if not check_type(obj, type):
            raise BadSignature()
    except SignatureExpired as e:
        # 410
        raise e
    except BadSignature as e:
        # 404
        raise e
    if 'u' not in obj:
        # 401 bad request
        return None
    return get_user_model().objects.filter(pk=obj['u']).first()


def verify_email_token(token, type, max_age):
    # BadSignatureかSinatureExpiredが例外として返される
    # それ以外は、userの情報が無い、userが存在しない場合はNoneを返し、いる場合はuserを返す
    try:
        obj = signing.loads(token, max_age=max_age)
        if not check_type(obj, type):
            raise BadSignature()
    except SignatureExpired as e:
        # 410
        raise e
    except BadSignature as e:
        # 404
        raise e
    if 'u' not in obj:
        # 401 bad request
        return None
    if 'email' not in obj:
        # 401 bad request
        return None
    return get_user_model().objects.filter(pk=obj['u']).first(), obj['email']


def get_user_and_tokentype(token):
    try:
        obj = signing.loads(token)
        if "type" not in obj or "u" not in obj:
            raise BadSignature()
    except BadSignature as e:
        # signatureが壊れてる
        raise e
    return get_user_model().objects.filter(pk=obj['u']).first(), obj["type"]


def get_email_from_token(token):
    try:
        obj = signing.loads(token)
        if "type" not in obj or "u" not in obj or "email" not in obj:
            raise BadSignature()
    except BadSignature as e:
        # signatureが壊れてる
        raise e
    return obj["email"]


def create_email_confirm_token(user):
    return create_token(user, TOKEN_EMAIL_CONFIRM)


def verify_email_confirm_token(token, max_age):
    return verify_token(token, TOKEN_EMAIL_CONFIRM, max_age)


def create_email_change_token(user, email):
    return create_email_token(user, email, TOKEN_EMAIL_CHANGE)


def verify_email_change_token(token, max_age):
    return verify_email_token(token, TOKEN_EMAIL_CHANGE, max_age)


def create_activation_token(user):
    return create_token(user, TOKEN_ACTIVATION)


def verify_activation_token(token, max_age):
    return verify_token(token, TOKEN_ACTIVATION, max_age)


def create_password_reset_token(user):
    return create_token(user, TOKEN_PASSWORD_RESET)


def verify_password_reset_token(token, max_age):
    return verify_token(token, TOKEN_PASSWORD_RESET, max_age)


def create_email_confirmation_url(token):
    return (api_settings.FRONTEND_DOMAIN_WITH_SCHEME + api_settings.EMAIL_CONFIRMATION_PATH_TEMPLATE)\
        .format(token=token)


def create_email_change_url(token):
    return (api_settings.FRONTEND_DOMAIN_WITH_SCHEME + api_settings.EMAIL_CHANGE_PATH_TEMPLATE)\
        .format(token=token)


def create_password_reset_url(token):
    return (api_settings.FRONTEND_DOMAIN_WITH_SCHEME + api_settings.PASSWORD_RESET_PATH_TEMPLATE).format(token=token)


def create_activation_url(token):
    return (api_settings.FRONTEND_DOMAIN_WITH_SCHEME + api_settings.ACTIVATION_PATH_TEMPLATE).format(token=token)



from django.utils.translation import ugettext as _
from requests.exceptions import HTTPError
from rest_framework import status
from rest_framework.generics import GenericAPIView
from rest_framework.permissions import AllowAny
from rest_framework.serializers import ValidationError
from social_core.backends.oauth import BaseOAuth2
from social_core.exceptions import MissingBackend
from social_django.utils import load_strategy, load_backend

from drf_jwt_util.views import create_jwt_response

from .serializers import SocialSerializer


class InvalidBackendException(Exception):
    pass


class SocialLoginView(GenericAPIView):
    serializer_class = SocialSerializer
    permission_classes = (AllowAny,)
    authentication_classes = ()

    def post(self, request):
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        provider = serializer.data.get('provider', None)
        strategy = load_strategy(request)
        try:
            backend = load_backend(strategy=strategy, name=provider,
                                   redirect_uri=None)
        except MissingBackend:
            raise ValidationError({"provider": _("invalid provider")})
        try:
            if isinstance(backend, BaseOAuth2):
                access_token = serializer.data.get('access_token')
                user = backend.do_auth(access_token)
            else:
                raise InvalidBackendException
        except HTTPError:
            raise ValidationError({"backend error"})
        except InvalidBackendException:
            raise ValidationError({"provider": _("invalid provider")})
        return create_jwt_response(request, user, statuscode=status.HTTP_200_OK)



# -*- coding: utf-8 -*-
from django.contrib.auth import get_user_model
from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers
from rest_framework_jwt.compat import PasswordField
from rest_framework_jwt.serializers import RefreshJSONWebTokenSerializer as OrigRefreshJSONWebTokenSerializer
from rest_framework_jwt.serializers import JSONWebTokenSerializer

UserModel = get_user_model()


def get_username_field():
    try:
        username_field = UserModel.USERNAME_FIELD
    except AttributeError:
        username_field = 'username'

    return username_field


def get_field_by_name(option, name):
    fields = {f.name: f for f in option.fields}
    if name not in fields:
        return serializers.CharField()()
    field = fields[name]
    if field.__class__ not in serializers.ModelSerializer.serializer_field_mapping:
        return serializers.CharField()
    return serializers.ModelSerializer.serializer_field_mapping[field.__class__]()


class ObtainJSONWebTokenSerializer(JSONWebTokenSerializer):
    def __init__(self, *args, **kwargs):
        super(JSONWebTokenSerializer, self).__init__(*args, **kwargs)
        self.fields[get_username_field()] = get_field_by_name(UserModel._meta, get_username_field())
        self.fields['password'] = PasswordField(write_only=True)


class RefreshJSONWebTokenSerializer(OrigRefreshJSONWebTokenSerializer):
    def _check_user(self, payload):
        uid = payload['uid']
        if not uid:
            msg = _('Invalid payload.')
            raise serializers.ValidationError(msg)
        try:
            user = get_user_model().objects.get(id=uid)
        except UserModel.DoesNotExist:
            msg = _("User doesn't exist.")
            raise serializers.ValidationError(msg)

        if not user.is_active:
            msg = _('User account is disabled.')
            raise serializers.ValidationError(msg)

        return user


class DefaultUserSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserModel
        fields = ["pk", "email", "username"]
        read_only_fields = [get_username_field()]


class ValidatePasswordMixin:
    def validate_password1(self, value):
        try:
            validate_password(value)
        except ValidationError as exc:
            raise serializers.ValidationError(str(exc))
        return value

    def validate(self, data):
        if data['password1'] != data['password2']:
            raise serializers.ValidationError({"password2": _("You must type the same email each time.")})
        self.password = data["password1"]
        del data["password1"]
        del data["password2"]
        return data


class RegisterSerializer(ValidatePasswordMixin, serializers.ModelSerializer):
    class Meta:
        model = get_user_model()
        fields = [get_username_field(), "password1", "password2"]
    password1 = serializers.CharField(write_only=True)
    password2 = serializers.CharField(write_only=True)

    def save(self, **kwargs):
        user = super(RegisterSerializer, self).save(**kwargs)
        user.set_password(self.password)
        user.save()
        return user


class PasswordSerializer(ValidatePasswordMixin, serializers.Serializer):
    password1 = serializers.CharField()
    password2 = serializers.CharField()


class EmailSerializer(serializers.Serializer):
    email = serializers.EmailField()



from setuptools import setup, find_packages

setup(
    name="drf_jwt_util",
    version="0.1.5",
    description="drf jwt util",
    author='Shunsuke Aihara',
    author_email='aihara@argmax.jp',
    url='https://github.com/shunsukeaihara/drf_jwt_util',
    packages=find_packages(),
    include_package_data=True,
    install_requires=["django", "djangorestframework", "djangorestframework-jwt", "social-auth-app-django"],
    tests_require=["tox", "factory_boy", "faker>=1.0.6"],
    license="MIT License",
)
